function F = fast_nondominated_sort(P)
% ni, the number of solutions which dominate the solution i
% Si, a set of solutions which the solution i dominates

PoolSize = size(P,1);
S = zeros(PoolSize, PoolSize);
n = zeros(PoolSize, 1);
F = zeros(PoolSize, 1);

for j = 1:PoolSize
    p = P(j,:);
    for k = 1:PoolSize
        q = P(k,:);
        if pareto_dominate(p, q) % if p dominates q then
            S(j, k) = 1; % include q in Sp
        elseif pareto_dominate(q, p) % if p is dominated by q then
            n(j, :) = n(j,:) + 1; % increment np
        end
    end
    if n(j, :) == 0 % if no solution dominates p then
        F(j, :) = 1; % p is a member of the first front
    end
end

i = 1;
temp_index = 1:PoolSize;
while sum(F == i) ~= 0
    H = zeros(PoolSize, 1);
    Fi = temp_index(F==i);
    for j = 1:length(Fi)
        p = Fi(j); % for each member p in Fi
        Sp = temp_index(S(p,:)==1);
        for k = 1:length(Sp)
            q = Sp(k); % modify each member from the set Sp
            n(q,:) = n(q,:) - 1; % decrement nq by one
            if n(q,:) == 0 
                H(q, :) = 1; % if nq is zero, q is a member of a list H
            end
        end
    end
    i = i + 1;
    F(logical(H),:) = i; % current front is formed with all members of H
end
end