%% Add path
addpath(genpath('MM_testfunctions/'));
addpath(genpath('Indicator_calculation/'));
addpath(genpath('Supporting_Functions/'));

%%
close all
clearvars
clearvars -global
clc

%% Parameters
[arch.problem, ch] = get_problem_name();
[cand_dim, n_obj] = dim_CEC19(arch.problem);
if ch(1) == 18 || strcmp(ch, 'MMF14_a') || ...
        ch(1) == 19 || strcmp(ch, 'MMF15_a') || ...
        ch(1) == 14 || strcmp(ch, 'MMF14') || ...
        ch(1) == 15 || strcmp(ch, 'MMF15')
    n_obj = input('Enter number of objectives:');
    if n_obj ~= 3 && n_obj ~= 5 && n_obj ~= 8 && n_obj ~= 10
        error('This code supports only 3, 5, 8 and 10 objective problems.')
    end
    cand_dim = n_obj;
end
[arch.XL, arch.XU] = dec_bounds_CEC19(arch.problem, cand_dim);

arch.size = 100 * cand_dim;
max_FES = 5000 * cand_dim;
n_dir = arch.size;

if n_obj == 2
    arch.ref_dir = get_2D_weight_vectors(n_dir);
elseif n_obj >= 3
    filename = ['Obj', num2str(n_obj), '.mat'];
    load(filename);
    arch.ref_dir = W;
    n_dir = size(W, 1);
    arch.size = n_dir;
end

arch.sp_size = arch.size / n_dir;
Gmax = floor(max_FES / n_dir);
arch.theta = 5;

arch.nbr_frac = 0.2;
arch.nbr_size = scientific_round(arch.nbr_frac * n_dir);
arch.lookup_nbr_dir = form_ref_table(arch.ref_dir, n_dir - 1);

arch.F_m = 0.5;
arch.CR_m = 0.2;
arch.eta_c_m = 30/50;
arch.F = 0.5;
arch.CR = 0.2;
arch.eta_c = 30;
arch.eta_m = 20;
arch.mut_prob = 1 / cand_dim;
arch.cross_prob = 1;
arch.frac = 0.75;  % mutation probability for mutation switching scheme

[arch.z_ide, arch.ref_pt] = HV_ref_pt_CEC19(arch.problem, n_obj);
norm_ref_vecs = [arch.z_ide; arch.ref_pt];

arch.dec = zeros(arch.size, cand_dim); 
arch.obj = zeros(arch.size, n_obj);
arch.obj_norm = zeros(arch.size, n_obj);
%% Population Initialization
for cand_i = 1:arch.size
    temp = initCand(cand_dim, arch.XL, arch.XU);
    arch.dec(cand_i, :) = temp;
    arch.obj(cand_i, :) = fitness(arch.problem, temp, n_obj);
end

arch.obj_norm = normalize_obj(arch.obj, norm_ref_vecs);
arch.assoc_dir = associate_points(arch.ref_dir, arch.obj_norm);
arch.scalar_fit = scalarization2(arch);
%% Optimization loop
for gen_i = 1 : Gmax
    flag = 0;
    if gen_i > scientific_round(0.5 * Gmax)
        flag = 1;
    end

    % Initialize params vec
    S_F = zeros(n_dir, 1);
    S_CR = zeros(n_dir, 1);
    S_eta_c = zeros(n_dir, 1);

    % For each direction
    for dir_i = 1 : n_dir
        cur_nbrs = zeros(1, n_dir);
        cnt = 0;
        j = 1;        
        % cur_nbrs: non-empty reference lines neighboring dir_i
        while (cnt < arch.nbr_size) && (j <= size(arch.lookup_nbr_dir, 2))
            if ~isempty(arch.dec(arch.assoc_dir == arch.lookup_nbr_dir(dir_i, j), :))
                cur_nbrs(1, j) = arch.lookup_nbr_dir(dir_i, j);
                cnt = cnt + 1;
            end
            j = j + 1;
        end

        % cur_dir: dir_i (if non-empty) or its random non-empty neighbor %%%%%%%%%%
        cur_dir = dir_i;
        while isempty(arch.dec(arch.assoc_dir == cur_dir))
            temp = cur_nbrs(cur_nbrs ~= 0);
            a = randi([1, length(temp)], 1);
            cur_dir = temp(a);
        end

        % current_pt: choose a random solution from cur_dir
        temp_idx = 1:arch.size;
        temp_idx = temp_idx(arch.assoc_dir == cur_dir);
        r1 = randi([1, length(temp_idx)], 1);
        current_pt = choose_candidate(arch, temp_idx(r1));

        % Variation / Reproduction of new candidate within cur_nbrs
        [arch.eta_c, arch.F, arch.CR] = get_mutate_params(arch.eta_c_m, 0.1, arch.F_m, 0.1, arch.CR_m, 0.1);
        arch.eta_c = 50 * arch.eta_c;
        if rand(1) > arch.frac
            new_pt = offspring_creation(arch, current_pt.dec, 'DE_mutate', cur_nbrs);
            new_pt = offspring_creation(arch, new_pt, 'Poly_mutate', cur_nbrs);
        else
            new_pt = offspring_creation(arch, current_pt.dec, 'SBX_mutate', cur_nbrs);
            new_pt = offspring_creation(arch, new_pt, 'Poly_mutate', cur_nbrs);
        end
        new_obj_vec = fitness(arch.problem, new_pt, size(arch.obj, 2)); 
        if n_obj == 2
            new_obj_vec = new_obj_vec';
        end
        if ch(1) == 16 || strcmp(ch, 'MMF1_e') || ...
            ch(1) == 17 || strcmp(ch, 'MMF1_z')
            new_obj_vec = new_obj_vec';
        end
        dom_stat = (pareto_dominate(current_pt.obj, new_obj_vec) ~= 1); % true, if current_pt doesn't dominate new_pt

        success = 0; % success = 0 => current_pt has not entered the archive
        if (dom_stat == 1) % if current_pt doesn't dominate new_pt
            success = 1; % success = 1 => current_pt enters archive

            % add new_pt to archive and evaluate associated params of new_pt
            arch.dec = vertcat(arch.dec, new_pt);
            arch.obj = vertcat(arch.obj, new_obj_vec);            
            arch.size = arch.size + 1;

            delta_vec = norm_ref_vecs(2, :) - norm_ref_vecs(1, :);
            delta_vec(1, delta_vec == 0) = eps;
            new_obj_norm = (new_obj_vec - norm_ref_vecs(1, :)) ./ delta_vec; %%%%%%%%%%%%
            new_assoc_dir = associate_points(arch.ref_dir, new_obj_norm);
            d2_val = d2(new_obj_norm, arch.ref_dir(new_assoc_dir, :)); 
            d1_val = norm(new_obj_norm .* arch.ref_dir(new_assoc_dir, :)) / norm(arch.ref_dir(new_assoc_dir, :));
            new_scalar_fit = d1_val + arch.theta * d2_val;

            arch.obj_norm = vertcat(arch.obj_norm, new_obj_norm);
            arch.assoc_dir = vertcat(arch.assoc_dir, new_assoc_dir);
            arch.scalar_fit = vertcat(arch.scalar_fit, new_scalar_fit);

            % perform filtering to maintain convergence and diversity
            arch = selection(arch, flag, n_obj);
        end
        if success == 1 % if new_pt enters archive, store the params
            S_eta_c(dir_i, 1) = arch.eta_c;
            S_F(dir_i, 1) = arch.F;
            S_CR(dir_i, 1) = arch.CR;
        end
    end

    % Update params
    if sum(S_eta_c) ~= 0
        S_eta_c = S_eta_c / 50;
        arch.eta_c_m = sum(S_eta_c) / length(S_eta_c(S_eta_c ~= 0));
    end
    if sum(S_F) ~= 0
        arch.F_m = sum(S_F) / length(S_F(S_F ~= 0));
    end
    if sum(S_CR) ~= 0
        arch.CR_m = sum(S_CR) / length(S_CR(S_CR ~= 0));
    end

    clc; disp([' Progress: ', num2str(gen_i * 100 / Gmax), '%']);
end

F = fast_nondominated_sort(arch.obj);
arch1 = choose_candidate(arch, F == 1);

%% Display results
arch1.perf = zeros(1, 4); % PSP, IGDX, HV, IGDF
arch1.perf = perf_ana2(arch1);
disp('=================');
disp(['Results for ',arch1.problem]);
disp('=================');
fprintf(1, 'rPSP: %0.4f\nIGDX:%0.4f\n', 1/arch1.perf(1,1), arch1.perf(1,2));
fprintf(1, 'rHV: %0.4f\nIGDF:%0.4f\n', 1/arch1.perf(1,3), arch1.perf(1,4));

%% Save results
fp = fopen('Results/Perf_LORD.txt', 'a+');
fprintf(fp, '%s\t', arch1.problem);
fprintf(fp, '%0.9f\t%0.9f\t', 1/arch1.perf(1,1), arch1.perf(1,2));
fprintf(fp, '%0.9f\t%0.9f\n', 1/arch1.perf(1,3), arch1.perf(1,4));
fclose(fp);
save(['Results/', arch1.problem, '.mat'], 'arch');

if n_obj == 2 || n_obj == 3
    visualize_MMOPs(arch1)
end